	;; Crunched Flipper 1.0 by Micro Digital
	;; Based on the stand-alone pucrunch decompressor by Pasi Ojala
	;; Optimised by Marko Mkel

	;; This file can be assembled with DASM 2.12.04.
	
	processor 6502
	seg code
	org $1001
	dc.w nextln
	dc.w 2002
	.byte $9e		; SYS
	.byte $30 + (start / 1000)
	.byte $30 + ((start - (start / 1000 * 1000)) / 100)
	.byte $30 + ((start - (start / 100 * 100)) / 10)
	.byte $30 + (start - (start / 10 * 10))
	.byte $3a		; :
	.byte $a2		; NEW
	.byte 0
nextln	.word 0

LZPOS	equ $9e		; 2 ZeroPage temporaries
table	equ $200	; RLE table
bitstr	equ $f7
esc	equ $f8
prgtype	equ $f9
OUTPOS	equ $fb


start	sei
	ldx #8
	stx $900f		; black screen
	lda #0
zero	sta $5c,x
	dex
	bpl zero
;	txs			; no need to initialize stack pointer to $ff
	ldx #block_stack_end-block_stack_
cpaux$	lda block_stack-1,x
	sta.wx block_stack_-1,x
	dex
	bne cpaux$
	ldy #>blocke_ - >block1_ + 1
reloc$	lda blocke_-256,x
	sta blocke-256,x
	dex
	bne reloc$
	dec reloc$+2
	dec reloc$+5
	dey
	bne reloc$

	jsr decruncher
	lda #3
	jsr decnext
	ldx #<lastpart
	lda #>lastpart
	jsr declast	; decrunch the main part
;	ldx #$40
;cp	lda $300-1,x
;	sta $218-1,x	; don't relocate these
;	dex
;	bne cp
	beq lastpart	; branch always

; getbits: Gets X bits from the stream
; ** Scratches X, returns the value in A **
getbits	jsr getbit
	rol
getchk	dex
getchkf	bne getbits
	clc
	rts

; getval : Gets a 'static huffman coded' value
; ** Scratches X, returns the value in A **
getval	inx		; X <- 1
	txa		; set the top bit (value is 1..255)
gv0	jsr getbit
	bcc getchk	; got 0-bit
	inx
mg	cpx #6		; ** PARAMETER unary code maximum length + 1
	bne gv0
	beq getchk	; inverse condition -> jump always

getbit	asl bitstr
	bne gbend
getnew	pha
INPOS = .+1
	lda block1	; ** PARAMETER
	;; xxxxxxxx 101yyyyy yyyyyyyy
	inc INPOS
	bne 0$
	inc INPOS+1
0$	rol		; Shift out the next bit and
			; shift in C=1 (last bit marker)
	sta bitstr	; bitstr initial value = $80 == empty
	pla
gbend	rts

	org $1078
	dc.b 5,2,3,7,4	; flipper colour table (originally at $1ffb)
lastpart
getbyt	jsr getnew
	lda bitstr
	ror
	rts

	;; decrunch last file
declast	stx OUTPOS
	;; decrunch next file
decnext	sta OUTPOS+1
	;; decrunch first file
decruncher
	jsr getbyt
	sta esc		; starting escape

	jsr getbyt	; read # of escape bits
	sta escB0+1
	sta escB1+1
	lda #8		; C preserved after getbyt
	sec
	sbc escB1+1
	sta noesc+1	; 8-escBits

;	jsr getbyt	; same value (6) for all three files (done with -m5)
;	sta mg+1	; maxGamma + 1
;	lda #9		; C preserved after getbyt
;	sec		; C=1 from last operation
;	sbc mg+1	; 8 - maxGamma == (8 + 1) - (maxGamma + 1)
;	sta longrle+1
;	jsr getbyt
;	sta mg1+1	; (1<<maxGamma)
;	asl
;	sbc #0		; C=0 after ASL
;	sta mg21+1	; (2<<maxGamma) - 1
;	jsr getbyt	; same value (0) for all three files
;	sta elzpb+1	; extraLZPosBits

	jsr getbyt	; rleUsed
	ldx #0
	tay
;	beq done$	; Y == 0 ? (not in this case)
cprle$	jsr getbyt
	sta table,x	; initialize the RLE table
	inx
	dey
	bne cprle$
done$	lda #$80
	sta bitstr
	jmp main

block_stack
	rorg $fa
block_stack_

putch	sta $9400	; ** parameter OUTPOS
	inc OUTPOS	; ZP
	bne 0$
	inc OUTPOS+1	; ZP
0$	dex
	rts
newesc	ldy esc		; remember the old code (top bits for escaped byte)
escB0	ldx #2		; ** PARAMETER	0..8
	jsr getchkf	; get & save the new escape code
	sta esc
	tya		; pre-set the bits
	; Fall through and get the rest of the bits.
noesc	ldx #6		; ** PARAMETER	8..0
	jsr getchkf
	jsr putch	; output the escaped/normal byte
	; Fall through and check the escape bits again
main	ldy #0		; Reset to a defined state
	tya		; A = 0
escB1	ldx #2		; ** PARAMETER	0..8
	jsr getchkf	; X = 0
	cmp esc
	bne noesc
	; Fall through to packed code

	jsr getval	; X = 0
	sta LZPOS	; xstore - save the length for a later time
	lsr		; cmp #1	; LEN == 2 ? (A is never 0)
	bne lz77	; LEN != 2	-> LZ77
	;tya		; A = 0

	jsr getbit	; X = 0
	bcc lz77_2	; A=0 -> LZPOS+1
	; e..e01
	jsr getbit	; X = 0
	bcc newesc	; e..e010
	; e..e011
	iny		; Y is 1 bigger than MSB loops
	jsr getval	; Y is 1, get len, X = 0
	sta LZPOS	; xstore - Save length LSB
mg1	cmp #32		; ** PARAMETER 63-64 -> C clear, 64-64 -> C set..
	bcc chrcode	; short RLE, get bytecode

longrle	ldx #3		; ** PARAMETER	111111xxxxxx
	jsr getbits	; get 3/2/1 more bits to get a full byte, X = 0
	sta LZPOS	; xstore - Save length LSB

	jsr getval	; length MSB, X = 0
	tay		; Y is 1 bigger than MSB loops

chrcode	jsr getval	; Byte Code, X = 0
	tax		; this is executed most of the time anyway
	lda table-1,x	; Saves one jump if done here (loses one txa)

	cpx #32		; 31-32 -> C clear, 32-32 -> C set..
	bcc 1$		; 1..31, we got the right byte from the table

	; Ranks 32..64 (11111xxxxx), get byte..
	txa		; get back the value (5 valid bits)
	ldx #3
	jsr getbits	; get 3 more bits to get a full byte, X = 0

1$	ldx LZPOS	; xstore - get length LSB
	inx		; adjust for cpx#$ff;bne -> bne
dorle	jsr putch	;+dex
	bne dorle	; xstore 0..255 -> 1..256
	dey
	bne dorle	; Y was 1 bigger than wanted originally
mainbeq	beq main	; reverse condition -> jump always

lz77	jsr getval	; X = 0
mg21	cmp #63		; ** PARAMETER	Clears carry (is maximum value)
	beq newesc-1	; end of file

	sbc #0		; C is clear -> subtract 1  (1..126 -> 0..125)
;elzpb	ldx #0		; ** PARAMETER (more bits to get)
;	jsr getchkf	; clears Carry, X = 0
	clc		; just clear carry (no more bits to get)

lz77_2	sta LZPOS+1	; offset MSB
	ldx #8
	jsr getbits	; clears Carry, X = 0
	; Note: Already eor:ed in the compressor..
	;eor #255	; offset LSB 2's complement -1 (i.e. -X = ~X+1)
	adc OUTPOS	; -offset -1 + curpos (C is clear)
	ldx LZPOS	; xstore = LZLEN (read before it's overwritten)

	sta LZPOS
	lda OUTPOS+1
	sbc LZPOS+1	; takes C into account
	sta LZPOS+1	; copy X+1 number of chars from LZPOS to OUTPOS
	;ldy #0		; Y was 0 originally, we don't change it

	inx		; adjust for cpx#$ff;bne -> bne
lzloop	lda (LZPOS),y	; using abs,y is 3 bytes longer, only 1 cycle/byte faster
	iny		; Y does not wrap because X=0..255 and Y initially 0
	jsr putch	;+dex
	bne lzloop	; X loops, (256,1..255)
	beq mainbeq

block_stack_end
	rend

	;; the crunched files
	;; created with
	;; pucrunch -c0 -m5 $in | \
	;; perl -e 'undef $/;$_=<>;s/^..pu..(.)..(.)...../$1$2/os;print' > $out
block1_
	;; These must be relocated, or otherwise the last part
	;; would overwrite the crunched bytes
	rorg $2000-3224
block1	incbin "crc"		; 9400-95ff (colours)
	incbin "cr1"		; 0300-03ff (graphics)
	incbin "cr2"		; 107d-1fff (main program)
blocke
	rend
blocke_
